"""API for exporting Slack channels"""

import datetime as dt
import logging
import logging.config
from pathlib import Path
import re
from typing import List, Optional

from babel import Locale
from babel.numbers import format_decimal
import pytz

from . import __version__
from . import settings
from .my_fpdf import MyFPDF
from .helpers import (
    transform_encoding,
    read_array_from_json_file,
    write_array_to_json_file,
)
from .locales import LocaleHelper
from .slack_service import SlackService
from .message_transformer import MessageTransformer

logging.config.dictConfig(settings.DEFAULT_LOGGING)
logger = logging.getLogger(__name__)


class SlackChannelExporter:
    """Class for exporting slack channels to PDF files

    This class will fetch all messages including threads from a Slack
    channel and then export them into a PDF file.

    The export will only include the text content, but not any media
    content like icons, images, etc. Media content is represented by
    placeholders with hyperlinks.

    Active elements of Slack message like buttons, menus, etc.
    are represented by placeholders.

    Emojis are included with their textual representation.

    Names of users, bots and channels are automatically resolved if possible.

    Attachments and blocks (sections only for now) are supported.

    """

    def __init__(
        self,
        slack_token: str,
        my_tz: pytz.BaseTzInfo = None,
        my_locale: Locale = None,
        add_debug_info: bool = False,
        logfile_path: Path = None,
    ):
        """
        Args:
            slack_token: OAuth token to be used for all calls to the Slack API
                "TEST" can be provided to run in test mode
            my_tz: override system's timezone
            my_locale: override system's default locale
            add_debug_info: wether to add debug info to message output

        """
        self._bot_names = dict()
        if slack_token is None:
            raise ValueError("slack_token can not be null")

        self._slack_service = SlackService(slack_token)

        # set locale & timezone
        author_info = self._slack_service.author_info()
        self._locale_helper = LocaleHelper(my_locale, my_tz, author_info)
        logger.info(
            "Locale is: %s [%s]",
            self._locale_helper.locale.get_display_name(),
            self._locale_helper.locale,
        )
        logger.info("Timezone is: %s", self._locale_helper.timezone)
        self._transformer = MessageTransformer(
            slack_service=self._slack_service,
            locale_helper=self._locale_helper,
            font_family_mono_default=settings.FONT_FAMILY_MONO_DEFAULT,
        )

        # validate add_debug_info
        if not isinstance(add_debug_info, bool):
            raise ValueError("add_debug_info must be bool")
        self._add_debug_info = add_debug_info
        if add_debug_info:
            logger.info("Adding DEBUG info to PDF")

        if logfile_path:
            pass

    def _parse_message_and_write_to_pdf(
        self,
        document: MyFPDF,
        msg: dict,
        margin_left: int,
        last_user_id: str,
        full_date: bool = False,
    ) -> Optional[str]:
        """parse a message and write it to the PDF"""

        if "user" in msg:
            user_id = msg["user"]
            is_bot = False
            if user_id in self._slack_service.user_names():
                user_name = self._slack_service.user_names()[user_id]
            else:
                user_name = f"unknown_user_{user_id}"

        elif "bot_id" in msg:
            user_id = msg["bot_id"]
            is_bot = True
            if "username" in msg:
                user_name = transform_encoding(msg["username"])
            elif user_id in self._bot_names:
                user_name = self._bot_names[user_id]
            else:
                user_name = f"unknown_bot_{user_id}"

        elif "subtype" in msg and msg["subtype"] == "file_comment":
            is_bot = False
            if "user" in msg["comment"]:
                user_id = msg["comment"]["user"]
                if user_id in self._slack_service.user_names():
                    user_name = self._slack_service.user_names()[user_id]
                else:
                    user_name = f"unknown_user_{user_id}"
            else:
                user_id = None
                user_name = None

        else:
            is_bot = False
            user_id = None
            user_name = None

        if user_name is not None:
            # start again on the left border
            document.set_left_margin(margin_left)
            document.set_x(margin_left)

            if last_user_id != user_id:
                # write user name and date only when user switches
                document.ln(settings.LINE_HEIGHT_SMALL)
                document.set_font(
                    settings.FONT_FAMILY_DEFAULT,
                    size=settings.FONT_SIZE_NORMAL,
                    style="B",
                )
                document.write(settings.LINE_HEIGHT_DEFAULT, user_name + " ")
                document.set_font(
                    settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_SMALL
                )
                if is_bot:
                    document.set_text_color(100, 100, 100)
                    document.write(settings.LINE_HEIGHT_DEFAULT, "App ")
                    document.set_text_color(0)

                datetime_str = (
                    self._locale_helper.get_datetime_formatted_str(msg["ts"])
                    if full_date
                    else self._locale_helper.get_time_formatted_str(msg["ts"])
                )
                document.write(settings.LINE_HEIGHT_DEFAULT, datetime_str)
                document.ln(settings.LINE_HEIGHT_DEFAULT)

            if "text" in msg and len(msg["text"]) > 0:
                text = msg["text"]
                if self._add_debug_info:
                    debug_text = (
                        f' [<s fontfamily="'
                        f'{settings.FONT_FAMILY_MONO_DEFAULT}" size="8">'
                        f'{msg["ts"]}]</s>'
                    )
                else:
                    debug_text = ""
                document.set_font(
                    settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_NORMAL
                )
                text_html = self._transformer.transform_text(
                    text, msg["mrkdwn"] if "mrkdwn" in msg else True
                )
                document.write_html(
                    settings.LINE_HEIGHT_DEFAULT, text_html + debug_text
                )
                document.ln(settings.LINE_HEIGHT_DEFAULT)

            if "reactions" in msg:
                # draw reactions
                for reaction in msg["reactions"]:
                    document.set_left_margin(margin_left + settings.TAB_WIDTH)
                    document.set_x(margin_left + settings.TAB_WIDTH)
                    document.set_font(
                        settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_NORMAL
                    )
                    document.write_html(
                        settings.LINE_HEIGHT_DEFAULT,
                        (
                            "["
                            + reaction["name"]
                            + "] ("
                            + str(reaction["count"])
                            + "):"
                        ),
                    )
                    document.ln()

                    # convert user IDs to names
                    users_with_names = list()
                    for user in reaction["users"]:
                        if user in self._slack_service.user_names():
                            user_name = self._slack_service.user_names()[user]
                        else:
                            user_name = "unknown_user_" + user

                        users_with_names.append("<b>" + user_name + "</b>")

                    document.set_left_margin(
                        margin_left + settings.TAB_WIDTH + settings.TAB_WIDTH
                    )
                    document.set_x(
                        margin_left + settings.TAB_WIDTH + settings.TAB_WIDTH
                    )
                    document.write_html(
                        settings.LINE_HEIGHT_DEFAULT, ", ".join(users_with_names)
                    )
                    document.ln()

                document.ln(settings.LINE_HEIGHT_SMALL)

            if "files" in msg:
                # draw files
                document.set_left_margin(margin_left + settings.TAB_WIDTH)
                document.set_x(margin_left + settings.TAB_WIDTH)

                for file in msg["files"]:
                    file_type = file.get("pretty_type", "")
                    file_name = file.get("name", "")
                    text = "[" + file_type + " file: <b>" + file_name + "</b>" + "]"
                    document.set_font(
                        settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_NORMAL
                    )
                    document.write_html(settings.LINE_HEIGHT_DEFAULT, text)
                    document.ln()

                    if "preview" in file:
                        text = file["preview"]
                        # remove document tag if any
                        match = re.match(r"<document>(.+)<\/document>", text)
                        if match is not None:
                            text = match.group(1)
                        # replace <p> with <br>
                        text = re.sub(r"<p>(.+)<\/p>", r"\1<br>", text)
                        # replace \r\n with <br>
                        text = re.sub(r"\n|\r\n", r"<br>", text)
                        # output
                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_NORMAL,
                        )
                        document.write_html(settings.LINE_HEIGHT_DEFAULT, text)
                        document.ln()

            if "attachments" in msg:
                # draw attachments
                document.set_left_margin(margin_left + settings.TAB_WIDTH)
                document.set_x(margin_left + settings.TAB_WIDTH)

                # draw normal text attachments
                for attach in msg["attachments"]:

                    if "mrkdwn_in" in attach:
                        mrkdwn_in = attach["mrkdwn_in"]
                    else:
                        mrkdwn_in = []

                    if "pretext" in attach:
                        document.set_left_margin(margin_left)
                        document.set_x(margin_left)
                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_NORMAL,
                        )
                        document.write_html(
                            settings.LINE_HEIGHT_DEFAULT,
                            self._transformer.transform_text(
                                attach["pretext"], "pretext" in mrkdwn_in
                            ),
                        )
                        document.set_left_margin(margin_left + settings.TAB_WIDTH)
                        document.set_x(margin_left + settings.TAB_WIDTH)
                        document.ln()

                    document.ln(settings.LINE_HEIGHT_SMALL)

                    if "author_name" in attach:
                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_LARGE,
                            style="B",
                        )
                        document.write(
                            settings.LINE_HEIGHT_DEFAULT,
                            self._transformer.transform_text(attach["author_name"]),
                        )
                        document.ln()

                    if "title" in attach:
                        title_text = self._transformer.transform_text(
                            attach["title"], "title" in mrkdwn_in
                        )

                        # add link to title if defined
                        if "title_link" in attach:
                            title_text = (
                                '<a href="'
                                + attach["title_link"]
                                + '">'
                                + title_text
                                + "</a>"
                            )

                        # add bold formatting to title
                        title_text = "<b>" + title_text + "</b>"

                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_NORMAL,
                        )
                        document.write_html(settings.LINE_HEIGHT_DEFAULT, title_text)
                        document.ln()

                    if "text" in attach:
                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_NORMAL,
                        )
                        document.write_html(
                            settings.LINE_HEIGHT_DEFAULT,
                            self._transformer.transform_text(
                                attach["text"], "text" in mrkdwn_in
                            ),
                        )
                        document.ln()

                    if "fields" in attach:
                        for field in attach["fields"]:
                            document.set_font(
                                settings.FONT_FAMILY_DEFAULT,
                                size=settings.FONT_SIZE_NORMAL,
                                style="B",
                            )
                            document.write(
                                settings.LINE_HEIGHT_DEFAULT,
                                self._transformer.transform_text(field["title"]),
                            )
                            document.ln()
                            document.set_font(
                                settings.FONT_FAMILY_DEFAULT,
                                size=settings.FONT_SIZE_NORMAL,
                            )
                            document.write_html(
                                settings.LINE_HEIGHT_DEFAULT,
                                self._transformer.transform_text(
                                    field["value"], "fields" in mrkdwn_in
                                ),
                            )
                            document.ln()

                    if "footer" in attach:
                        if "ts" in attach:
                            text = (
                                self._transformer.transform_text(attach["footer"])
                                + "|"
                                + self._locale_helper.get_datetime_formatted_str(
                                    attach["ts"]
                                )
                            )
                        else:
                            text = self._transformer.transform_text(attach["footer"])

                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_SMALL,
                        )
                        document.write(settings.LINE_HEIGHT_DEFAULT, text)
                        document.ln()

                    if "image_url" in attach:
                        image_url_html = (
                            '<a href="' + attach["image_url"] + '">[Image]</a>'
                        )
                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_NORMAL,
                        )
                        document.write_html(
                            settings.LINE_HEIGHT_DEFAULT, image_url_html
                        )
                        document.ln()

                    # action attachments
                    if "actions" in attach:
                        for action in attach["actions"]:
                            document.set_font(
                                settings.FONT_FAMILY_DEFAULT,
                                size=settings.FONT_SIZE_SMALL,
                            )
                            document.write_html(
                                settings.LINE_HEIGHT_DEFAULT,
                                (
                                    "["
                                    + self._transformer.transform_text(action["text"])
                                    + "] "
                                ),
                            )

                        document.ln()

                document.ln(settings.LINE_HEIGHT_SMALL)

            if "blocks" in msg:
                document.set_left_margin(margin_left + settings.TAB_WIDTH)
                document.set_x(margin_left + settings.TAB_WIDTH)

                for layout_block in msg["blocks"]:
                    block_type = layout_block["type"]
                    document.ln(settings.LINE_HEIGHT_SMALL)

                    # section layout blocks
                    if block_type == "section":
                        document.set_font(
                            settings.FONT_FAMILY_DEFAULT,
                            size=settings.FONT_SIZE_NORMAL,
                        )
                        document.write_html(
                            settings.LINE_HEIGHT_DEFAULT,
                            self._transformer.transform_text(
                                layout_block["text"]["text"],
                                layout_block["text"]["type"] == "mrkdwn",
                            ),
                        )
                        document.ln()

                        if "fields" in layout_block:
                            for field in layout_block["fields"]:
                                document.set_font(
                                    settings.FONT_FAMILY_DEFAULT,
                                    size=settings.FONT_SIZE_NORMAL,
                                )
                                document.write_html(
                                    settings.LINE_HEIGHT_DEFAULT,
                                    self._transformer.transform_text(
                                        field["text"], field["type"] == "mrkdwn"
                                    ),
                                )
                                document.ln()

                document.ln(settings.LINE_HEIGHT_SMALL)

        else:
            user_id = None
            logger.warning("Can not process message with ts %s", msg["ts"])
            document.write(
                settings.LINE_HEIGHT_DEFAULT, "[Can not process this message]"
            )
            document.ln()

        return user_id

    def _write_messages_to_pdf(
        self, document: MyFPDF, messages: List[dict], threads: List[dict]
    ) -> None:
        """writes messages with their threads to the PDF document"""
        last_user_id = None
        last_dt = None
        last_page = None

        if len(messages) > 0:
            messages = sorted(messages, key=lambda k: k["ts"])
            for msg in messages:

                msg_dt = self._locale_helper.get_datetime_from_ts(msg["ts"])

                # repeat user name for if last post from same user is older
                if last_dt is not None:
                    dt_delta = msg_dt - last_dt
                    minutes_delta = dt_delta / dt.timedelta(minutes=1)
                    if minutes_delta > settings.MINUTES_UNTIL_USERNAME_REPEATS:
                        last_user_id = None

                # write day seperator if needed
                if last_dt is None or msg_dt.date() != last_dt.date():
                    document.ln(settings.LINE_HEIGHT_SMALL)
                    document.ln(settings.LINE_HEIGHT_SMALL)
                    document.set_font(
                        settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_NORMAL
                    )

                    # draw divider line for next day
                    page_width = document.fw - 2 * settings.MARGIN_LEFT
                    x1 = settings.MARGIN_LEFT
                    x2 = x1 + page_width
                    y1 = document.get_y() + 3
                    document.line(x1, y1, x2, y1)

                    # stamp date on divider
                    date_text = self._locale_helper.format_date_full_str(msg_dt)
                    text_width = document.get_string_width(date_text)
                    x3 = (x2 - x1) / 2 + x1
                    x4 = x3 - (text_width / 2)
                    border_x = 3
                    document.set_fill_color(255, 255, 255)
                    document.set_x(x4 - border_x)
                    document.cell(
                        text_width + 2 * border_x,
                        settings.LINE_HEIGHT_DEFAULT,
                        date_text,
                        0,
                        0,
                        "C",
                        True,
                    )
                    document.ln()
                    last_user_id = None  # repeat user name for new day

                # repeat user name for new page
                if last_page != document.page_no():
                    last_user_id = None
                    last_page = document.page_no()

                last_user_id = self._parse_message_and_write_to_pdf(
                    document, msg, settings.MARGIN_LEFT, last_user_id
                )
                if "thread_ts" in msg and msg["thread_ts"] == msg["ts"]:
                    thread_ts = msg["thread_ts"]

                    if thread_ts in threads:
                        thread_messages = threads[thread_ts]
                        last_user_id = None
                        last_dt = None
                        thread_messages = sorted(thread_messages, key=lambda k: k["ts"])
                        for thread_msg in thread_messages:
                            if thread_msg["ts"] != thread_msg["thread_ts"]:
                                # repeat user name for if last post from same user is older
                                msg_dt = self._locale_helper.get_datetime_from_ts(
                                    thread_msg["ts"]
                                )
                                if last_dt is not None:
                                    dt_delta = msg_dt - last_dt
                                    minutes_delta = dt_delta / dt.timedelta(minutes=1)
                                    if (
                                        minutes_delta
                                        > settings.MINUTES_UNTIL_USERNAME_REPEATS
                                    ):
                                        last_user_id = None
                                last_dt = msg_dt
                                last_user_id = self._parse_message_and_write_to_pdf(
                                    document,
                                    thread_msg,
                                    settings.MARGIN_LEFT + settings.TAB_WIDTH,
                                    last_user_id,
                                    full_date=True,
                                )

                    last_user_id = None

                last_dt = msg_dt
        else:
            document.set_font(
                settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_NORMAL
            )
            document.ln()
            document.write(settings.LINE_HEIGHT_DEFAULT, "This channel is empty", "I")

    def run(
        self,
        channel_inputs: list,
        dest_path: Path = None,
        oldest: dt.datetime = None,
        latest: dt.datetime = None,
        page_orientation: str = "portrait",
        page_format: str = "a4",
        max_messages: int = None,
        write_raw_data: bool = False,
    ) -> dict:
        """Exports all message from a channel and stores them in a PDF

        Args:
        - channel_inputs: list of names and/or IDs of channels to retrieve messages from
        - dest_path: path to write output files to. Will use current working directory if None
        - oldest: oldest message to fetch in UNIX epoch
        - latest: latest message to fetch in UNIX epoch
        - page_orientation: orientation of pages  as defined in FPDF class
        - page_format: format of pages, see as defined in FPDF class
        - max_messages: maximum number of messages to retrieve
        - write_raw_data: will safe data received from API to files if true

        Returns:
        - info about export result
        """
        # set defaults
        success = False

        # input validation
        if max_messages is not None:
            if not isinstance(max_messages, int):
                raise TypeError("max_messages must be of type int")
        else:
            max_messages = settings.MAX_MESSAGES_PER_CHANNEL

        if oldest is not None:
            if not isinstance(oldest, dt.datetime):
                raise TypeError("oldest must be a dt.datetime")
            oldest = self._locale_helper.timezone.localize(oldest)

        if latest is not None:
            if not isinstance(latest, dt.datetime):
                raise TypeError("latest must be a dt.datetime")
            latest = self._locale_helper.timezone.localize(latest)

        if oldest is not None and latest is not None and oldest > latest:
            raise RuntimeError("ERROR: oldest has to be before latest")

        if oldest is not None or latest is not None:
            text = "Fetching messages"
            if oldest is not None:
                text += f" after {self._locale_helper.format_datetime_str(oldest)}"
                if latest is not None:
                    text += " and"

            if latest is not None:
                text += f" before {self._locale_helper.format_datetime_str(latest)}"
            logger.info(text)

        if not isinstance(channel_inputs, list):
            raise TypeError("channel_inputs must be of type list")

        # set destination path as current dir if not set
        # or check if given path exists
        if dest_path is None:
            dest_path = Path.cwd()
        else:
            if not isinstance(dest_path, Path):
                raise TypeError("dest_path must be of type Path")
            if not dest_path.is_dir():
                raise RuntimeError(
                    f"ERROR: give destination path does not exist: {dest_path}"
                )

        logger.info("Writing output to: %s", dest_path.absolute())

        if not isinstance(page_orientation, str):
            raise TypeError("page_orientation must be of type str")
        else:
            logger.info("Page orientation: %s", page_orientation.title())

        if not isinstance(page_format, str):
            raise TypeError("page_format must be of type str")
        else:
            logger.info("Page format: %s", page_format.title())

        if write_raw_data is not None and not isinstance(write_raw_data, bool):
            raise TypeError("write_raw_data must be of type bool")

        # prepare to process channels
        team_name = self._slack_service.team
        response = {"ok": success, "channels": dict(), "team_name": team_name}
        channel_count = 0
        success = True

        # process each channel
        for channel_input in channel_inputs:
            success_channel = False
            channel_count += 1
            if channel_input.upper() in self._slack_service.channel_names():
                channel_id = channel_input.upper()
            else:
                # flip channel_names since channel names are unique
                channel_names_ids = {
                    v: k for k, v in self._slack_service.channel_names().items()
                }
                if channel_input.lower() not in channel_names_ids:
                    logger.error(
                        "(%d/%d) Unknown channel '%s' on %s",
                        channel_count,
                        len(channel_inputs),
                        channel_input,
                        team_name,
                    )
                    continue
                else:
                    channel_id = channel_names_ids[channel_input.lower()]

            channel_name = self._slack_service.channel_names()[channel_id]
            filename_base = re.sub(r"[^\w\-_\.]", "_", team_name)
            filename_base_channel = filename_base + "_" + channel_name

            # fetch messages
            # if we have a client fetch data from Slack
            if not self._slack_service.is_test_mode:
                progress_str = (
                    f"({channel_count}/{len(channel_inputs)})"
                    if len(channel_inputs) > 1
                    else ""
                )
                logger.info("Current channel %s: %s", progress_str, channel_name)
                messages = self._slack_service.fetch_messages_from_channel(
                    channel_id, max_messages, oldest, latest
                )
                threads = self._slack_service.fetch_threads_from_messages(
                    channel_id, messages, max_messages, oldest, latest
                )
                self._bot_names = self._slack_service.fetch_bot_names_for_messages(
                    messages, threads
                )

                if write_raw_data:
                    # write raw data received from Slack API to file
                    write_array_to_json_file(
                        self._slack_service.user_names(),
                        dest_path / (filename_base + "_users"),
                    )
                    write_array_to_json_file(
                        self._bot_names, dest_path / (filename_base + "_bots")
                    )
                    write_array_to_json_file(
                        self._slack_service.channel_names(),
                        dest_path / (filename_base + "_channels"),
                    )
                    write_array_to_json_file(
                        self._slack_service.user_names(),
                        dest_path / (filename_base + "_usergroups"),
                    )
                    write_array_to_json_file(
                        messages, dest_path / (filename_base_channel + "_messages")
                    )
                    if len(threads) > 0:
                        write_array_to_json_file(
                            threads, dest_path / (filename_base_channel + "_threads")
                        )
            else:
                # if we don't have a client we will try to fetch from a file
                # this is used for testing
                messages = read_array_from_json_file(
                    dest_path / (filename_base_channel + "_messages")
                )
                threads = read_array_from_json_file(
                    filepath=dest_path / (filename_base_channel + "_threads"),
                    quiet=True,
                )

            # create PDF
            document = MyFPDF(
                page_orientation, settings.PAGE_UNITS_DEFAULT, page_format
            )

            # add all fonts to support unicode
            document.add_font(
                settings.FONT_FAMILY_DEFAULT,
                style="",
                fname="NotoSans-Regular.ttf",
                uni=True,
            )
            document.add_font(
                settings.FONT_FAMILY_DEFAULT,
                style="B",
                fname="NotoSans-Bold.ttf",
                uni=True,
            )
            document.add_font(
                settings.FONT_FAMILY_DEFAULT,
                style="I",
                fname="NotoSans-Italic.ttf",
                uni=True,
            )
            document.add_font(
                settings.FONT_FAMILY_DEFAULT,
                style="BI",
                fname="NotoSans-BoldItalic.ttf",
                uni=True,
            )
            document.add_font(
                settings.FONT_FAMILY_MONO_DEFAULT,
                style="",
                fname="NotoSansMono-Regular.ttf",
                uni=True,
            )
            document.add_font(
                settings.FONT_FAMILY_MONO_DEFAULT,
                style="B",
                fname="NotoSansMono-Bold.ttf",
                uni=True,
            )
            document.alias_nb_pages()
            document.add_page()

            # compile all values
            creation_date = dt.datetime.now(tz=self._locale_helper.timezone)
            creation_datetime_str = self._locale_helper.format_datetime_str(
                creation_date
            )

            # count all messages including threads
            message_count = len(messages)
            if len(threads) > 0:
                for _, thread_messages in threads.items():
                    message_count += len(thread_messages) - 1

            if message_count > 0:
                # find start and end date based on messages
                ts_extract = [d["ts"] for d in messages]
                ts_min = min(float(s) for s in ts_extract)
                ts_max = max(float(s) for s in ts_extract)

                start_date = self._locale_helper.get_datetime_from_ts(ts_min)
                start_date_str = self._locale_helper.format_datetime_str(start_date)
                end_date = self._locale_helper.get_datetime_from_ts(ts_max)
                end_date_str = self._locale_helper.format_datetime_str(end_date)
            else:
                start_date = None
                start_date_str = ""
                end_date = None
                end_date_str = ""

            # set variables for title, header, footer
            title = team_name + " / " + channel_name
            sub_title = "Slack channel export"
            page_title = title

            # set properties for document info
            document.set_author(self._slack_service.author)
            document.set_creator(f"Channel Export v{__version__}")
            document.set_title(title)
            # document.set_creation_date(creation_date)
            document.set_subject(sub_title)
            document.page_title = page_title

            # write title on first page
            document.set_font(
                settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_LARGE, style="B"
            )
            document.cell(0, 0, title, 0, 1, "C")
            document.ln(settings.LINE_HEIGHT_DEFAULT)

            document.set_font(
                settings.FONT_FAMILY_DEFAULT,
                size=settings.FONT_SIZE_NORMAL,
                style="B",
            )
            document.cell(0, 0, sub_title, 0, 1, "C")
            document.ln(settings.LINE_HEIGHT_DEFAULT)

            # write info block after title
            thread_count = len(threads.keys()) if len(threads) > 0 else 0
            export_infos = {
                "Slack workspace": team_name,
                "Channel": channel_name,
                "Exported at": creation_datetime_str,
                "Exported by": self._slack_service.author,
                "Start date": start_date_str,
                "End date": end_date_str,
                "Timezone": self._locale_helper.timezone,
                "Locale": f"{self._locale_helper.locale.get_display_name()}",
                "Messages": format_decimal(
                    message_count, locale=self._locale_helper.locale
                ),
                "Threads": format_decimal(
                    thread_count, locale=self._locale_helper.locale
                ),
                "Pages": "{nb}",
            }
            document.write_info_table(export_infos)
            document.add_page()

            # write messages to PDF
            self._write_messages_to_pdf(document, messages, threads)

            # store PDF
            filename_pdf = dest_path / (filename_base_channel + ".pdf")
            logger.info("Writing PDF file: %s", filename_pdf)
            try:
                document.output(str(filename_pdf))
                success_channel = True
            except IOError:
                logger.error("Failed to write PDF file:", exc_info=True)

            # compile response dict
            response["channels"][channel_id] = {
                "ok": success_channel,
                "channel_id": channel_id,
                "channel_name": channel_name,
                "filename_pdf": str(filename_pdf),
                "filename_base_channel": str(dest_path / filename_base_channel),
                "dest_path": str(dest_path),
                "page_format": page_format,
                "page_orientation": page_orientation,
                "max_messages": max_messages,
                "messages_total": max_messages,
                "export_infos": export_infos,
                "message_count": message_count,
                "thread_count": thread_count,
                "creation_date": creation_date,
                "start_date": start_date,
                "end_date": end_date,
                "timezone": self._locale_helper.timezone,
                "locale": self._locale_helper.locale,
            }
            success = success and success_channel

        response["ok"] = success
        return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"Special module to handle differences between Python 2 and 3 versions"

import sys

PY3K = sys.version_info >= (3, 0)

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen

try:
    from io import BytesIO
except ImportError:
    try:
        from cStringIO import StringIO as BytesIO
    except ImportError:
        from StringIO import StringIO as BytesIO

try:
    from hashlib import md5
except ImportError:
    try:
        from md5 import md5
    except ImportError:
        md5 = None


def hashpath(fn):
    h = md5()
    if PY3K:
        h.update(fn.encode("UTF-8"))
    else:
        h.update(fn)
    return h.hexdigest()


# Check if PIL is available (tries importing both pypi version and corrected or manually installed versions).
# Necessary for JPEG and GIF support.
# TODO: Pillow support
try:
    from PIL import Image
except ImportError:
    try:
        import Image
    except ImportError:
        Image = None

try:
    from HTMLParser import HTMLParser
except ImportError:
    from html.parser import HTMLParser

if PY3K:
    basestring = str
    unicode = str
    ord = lambda x: x
else:
    basestring = basestring
    unicode = unicode
    ord = ord

# shortcut to bytes conversion (b prefix)
def b(s):
    if isinstance(s, basestring):
        return s.encode("latin1")
    elif isinstance(s, int):
        if PY3K:
            return bytes([s])  # http://bugs.python.org/issue4588
        else:
            return chr(s)


def exception():
    "Return the current the exception instance currently being handled"
    # this is needed to support Python 2.5 that lacks "as" syntax
    return sys.exc_info()[1]



import os
from setuptools import find_packages, setup

from slackchannel2pdf import __version__


# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="slackchannel2pdf",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    description=(
        "slackchannel2pdf is a command line tool for exporting the text "
        "contents of any Slack channel to a PDF file"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Erik Kalkoken",
    author_email="kalkoken87@gmail.com",
    classifiers=[
        "Environment :: Console",
        "Environment :: Win32 (MS Windows)",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=[
        "Babel==2.7.0",
        "pytz==2019.1",
        "python-dateutil==2.8.0",
        "slackclient==2.1.0",
        "tzlocal==2.0.0",
    ],
    extras_require={"testing": ["PyPDF2", "coverage"]},
    entry_points={
        "console_scripts": [
            "slackchannel2pdf=slackchannel2pdf.cli:main",
        ],
    },
)



# ******************************************************************************
# TTFontFile class
#
# This class is based on The ReportLab Open Source PDF library
# written in Python - http://www.reportlab.com/software/opensource/
# together with ideas from the OpenOffice source code and others.
#
# Version:  1.04
# Date:     2011-09-18
# Author:   Ian Back <ianb@bpm1.com>
# License:  LGPL
# Copyright (c) Ian Back, 2010
# Ported to Python 2.7 by Mariano Reingart (reingart@gmail.com) on 2012
# This header must be retained in any redistribution or
# modification of the file.
#
# ******************************************************************************

from __future__ import with_statement

from struct import pack, unpack, unpack_from
import re
import warnings
from .php import die, substr, str_repeat, str_pad, strlen, count
from .py3k import b, ord


# Define the value used in the "head" table of a created TTF file
# 0x74727565 "true" for Mac
# 0x00010000 for Windows
# Either seems to work for a font embedded in a PDF file
# when read by Adobe Reader on a Windows PC(!)
_TTF_MAC_HEADER = False


# TrueType Font Glyph operators
GF_WORDS = 1 << 0
GF_SCALE = 1 << 3
GF_MORE = 1 << 5
GF_XYSCALE = 1 << 6
GF_TWOBYTWO = 1 << 7


def sub32(x, y):
    xlo = x[1]
    xhi = x[0]
    ylo = y[1]
    yhi = y[0]
    if ylo > xlo:
        xlo += 1 << 16
        yhi += 1
    reslo = xlo - ylo
    if yhi > xhi:
        xhi += 1 << 16
    reshi = xhi - yhi
    reshi = reshi & 0xFFFF
    return (reshi, reslo)


def calcChecksum(data):
    if strlen(data) % 4:
        data += str_repeat(b("\0"), (4 - (len(data) % 4)))
    hi = 0x0000
    lo = 0x0000
    for i in range(0, len(data), 4):
        hi += (ord(data[i]) << 8) + ord(data[i + 1])
        lo += (ord(data[i + 2]) << 8) + ord(data[i + 3])
        hi += lo >> 16
        lo = lo & 0xFFFF
        hi = hi & 0xFFFF
    return (hi, lo)


class TTFontFile:
    def __init__(self):
        self.maxStrLenRead = 200000  # Maximum size of glyf table to read in as string (otherwise reads each glyph from file)

    def getMetrics(self, file):
        self.filename = file
        with open(file, "rb") as self.fh:
            self._pos = 0
            self.charWidths = []
            self.glyphPos = {}
            self.charToGlyph = {}
            self.tables = {}
            self.otables = {}
            self.ascent = 0
            self.descent = 0
            self.TTCFonts = {}
            self.version = version = self.read_ulong()
            if version == 0x4F54544F:
                die("Postscript outlines are not supported")
            if version == 0x74746366:
                die("ERROR - TrueType Fonts Collections not supported")
            if version not in (0x00010000, 0x74727565):
                die("Not a TrueType font: version=" + str(version))
            self.readTableDirectory()
            self.extractInfo()

    def readTableDirectory(
        self,
    ):
        self.numTables = self.read_ushort()
        self.searchRange = self.read_ushort()
        self.entrySelector = self.read_ushort()
        self.rangeShift = self.read_ushort()
        self.tables = {}
        for i in range(self.numTables):
            record = {}
            record["tag"] = self.read_tag()
            record["checksum"] = (self.read_ushort(), self.read_ushort())
            record["offset"] = self.read_ulong()
            record["length"] = self.read_ulong()
            self.tables[record["tag"]] = record

    def get_table_pos(self, tag):
        offset = self.tables[tag]["offset"]
        length = self.tables[tag]["length"]
        return (offset, length)

    def seek(self, pos):
        self._pos = pos
        self.fh.seek(self._pos)

    def skip(self, delta):
        self._pos = self._pos + delta
        self.fh.seek(self._pos)

    def seek_table(self, tag, offset_in_table=0):
        tpos = self.get_table_pos(tag)
        self._pos = tpos[0] + offset_in_table
        self.fh.seek(self._pos)
        return self._pos

    def read_tag(self):
        self._pos += 4
        return self.fh.read(4).decode("latin1")

    def read_short(self):
        self._pos += 2
        s = self.fh.read(2)
        a = (ord(s[0]) << 8) + ord(s[1])
        if a & (1 << 15):
            a = a - (1 << 16)
        return a

    def unpack_short(self, s):
        a = (ord(s[0]) << 8) + ord(s[1])
        if a & (1 << 15):
            a = a - (1 << 16)
        return a

    def read_ushort(self):
        self._pos += 2
        s = self.fh.read(2)
        return (ord(s[0]) << 8) + ord(s[1])

    def read_ulong(self):
        self._pos += 4
        s = self.fh.read(4)
        # if large uInt32 as an integer, PHP converts it to -ve
        return (
            (ord(s[0]) * 16777216) + (ord(s[1]) << 16) + (ord(s[2]) << 8) + ord(s[3])
        )  #     16777216  = 1<<24

    def get_ushort(self, pos):
        self.fh.seek(pos)
        s = self.fh.read(2)
        return (ord(s[0]) << 8) + ord(s[1])

    def get_ulong(self, pos):
        self.fh.seek(pos)
        s = self.fh.read(4)
        # iF large uInt32 as an integer, PHP converts it to -ve
        return (
            (ord(s[0]) * 16777216) + (ord(s[1]) << 16) + (ord(s[2]) << 8) + ord(s[3])
        )  #     16777216  = 1<<24

    def pack_short(self, val):
        if val < 0:
            val = abs(val)
            val = ~val
            val += 1
        return pack(">H", val)

    def splice(self, stream, offset, value):
        return (
            substr(stream, 0, offset) + value + substr(stream, offset + strlen(value))
        )

    def _set_ushort(self, stream, offset, value):
        up = pack(">H", value)
        return self.splice(stream, offset, up)

    def _set_short(self, stream, offset, val):
        if val < 0:
            val = abs(val)
            val = ~val
            val += 1
        up = pack(">H", val)
        return self.splice(stream, offset, up)

    def get_chunk(self, pos, length):
        self.fh.seek(pos)
        if length < 1:
            return ""
        return self.fh.read(length)

    def get_table(self, tag):
        (pos, length) = self.get_table_pos(tag)
        if length == 0:
            die("Truetype font (" + self.filename + "): error reading table: " + tag)
        self.fh.seek(pos)
        return self.fh.read(length)

    def add(self, tag, data):
        if tag == "head":
            data = self.splice(data, 8, b("\0\0\0\0"))
        self.otables[tag] = data

    ############################################/
    ############################################/

    ############################################/

    def extractInfo(self):
        #################/
        # name - Naming table
        #################/
        self.sFamilyClass = 0
        self.sFamilySubClass = 0

        name_offset = self.seek_table("name")
        format = self.read_ushort()
        if format != 0:
            die("Unknown name table format " + format)
        numRecords = self.read_ushort()
        string_data_offset = name_offset + self.read_ushort()
        names = {1: "", 2: "", 3: "", 4: "", 6: ""}
        K = list(names.keys())
        nameCount = len(names)
        for i in range(numRecords):
            platformId = self.read_ushort()
            encodingId = self.read_ushort()
            languageId = self.read_ushort()
            nameId = self.read_ushort()
            length = self.read_ushort()
            offset = self.read_ushort()
            if nameId not in K:
                continue
            N = ""
            if (
                platformId == 3 and encodingId == 1 and languageId == 0x409
            ):  # Microsoft, Unicode, US English, PS Name
                opos = self._pos
                self.seek(string_data_offset + offset)
                if length % 2 != 0:
                    die("PostScript name is UTF-16BE string of odd length")
                length //= 2
                N = ""
                while length > 0:
                    char = self.read_ushort()
                    N += chr(char)
                    length -= 1
                self._pos = opos
                self.seek(opos)

            elif (
                platformId == 1 and encodingId == 0 and languageId == 0
            ):  # Macintosh, Roman, English, PS Name
                opos = self._pos
                N = self.get_chunk(string_data_offset + offset, length).decode("latin1")
                self._pos = opos
                self.seek(opos)

            if N and names[nameId] == "":
                names[nameId] = N
                nameCount -= 1
                if nameCount == 0:
                    break

        if names[6]:
            psName = names[6]
        elif names[4]:
            psName = re.sub(" ", "-", names[4])
        elif names[1]:
            psName = re.sub(" ", "-", names[1])
        else:
            psName = ""
        if not psName:
            die("Could not find PostScript font name")
        self.name = psName
        if names[1]:
            self.familyName = names[1]
        else:
            self.familyName = psName
        if names[2]:
            self.styleName = names[2]
        else:
            self.styleName = "Regular"
        if names[4]:
            self.fullName = names[4]
        else:
            self.fullName = psName
        if names[3]:
            self.uniqueFontID = names[3]
        else:
            self.uniqueFontID = psName
        if names[6]:
            self.fullName = names[6]

        #################/
        # head - Font header table
        #################/
        self.seek_table("head")
        self.skip(18)
        self.unitsPerEm = unitsPerEm = self.read_ushort()
        scale = 1000 / float(unitsPerEm)
        self.skip(16)
        xMin = self.read_short()
        yMin = self.read_short()
        xMax = self.read_short()
        yMax = self.read_short()
        self.bbox = [(xMin * scale), (yMin * scale), (xMax * scale), (yMax * scale)]
        self.skip(3 * 2)
        indexToLocFormat = self.read_ushort()
        glyphDataFormat = self.read_ushort()
        if glyphDataFormat != 0:
            die("Unknown glyph data format " + glyphDataFormat)

        #################/
        # hhea metrics table
        #################/
        # ttf2t1 seems to use this value rather than the one in OS/2 - so put in for compatibility
        if "hhea" in self.tables:
            self.seek_table("hhea")
            self.skip(4)
            hheaAscender = self.read_short()
            hheaDescender = self.read_short()
            self.ascent = hheaAscender * scale
            self.descent = hheaDescender * scale

        #################/
        # OS/2 - OS/2 and Windows metrics table
        #################/
        if "OS/2" in self.tables:
            self.seek_table("OS/2")
            version = self.read_ushort()
            self.skip(2)
            usWeightClass = self.read_ushort()
            self.skip(2)
            fsType = self.read_ushort()
            if fsType == 0x0002 or (fsType & 0x0300) != 0:
                die(
                    "ERROR - Font file "
                    + self.filename
                    + " cannot be embedded due to copyright restrictions."
                )
                self.restrictedUse = True

            self.skip(20)
            sF = self.read_short()
            self.sFamilyClass = sF >> 8
            self.sFamilySubClass = sF & 0xFF
            self._pos += 10  # PANOSE = 10 byte length
            panose = self.fh.read(10)
            self.skip(26)
            sTypoAscender = self.read_short()
            sTypoDescender = self.read_short()
            if not self.ascent:
                self.ascent = sTypoAscender * scale
            if not self.descent:
                self.descent = sTypoDescender * scale
            if version > 1:
                self.skip(16)
                sCapHeight = self.read_short()
                self.capHeight = sCapHeight * scale
            else:
                self.capHeight = self.ascent

        else:
            usWeightClass = 500
            if not self.ascent:
                self.ascent = yMax * scale
            if not self.descent:
                self.descent = yMin * scale
            self.capHeight = self.ascent

        self.stemV = 50 + int(pow((usWeightClass / 65.0), 2))

        #################/
        # post - PostScript table
        #################/
        self.seek_table("post")
        self.skip(4)
        self.italicAngle = self.read_short() + self.read_ushort() / 65536.0
        self.underlinePosition = self.read_short() * scale
        self.underlineThickness = self.read_short() * scale
        isFixedPitch = self.read_ulong()

        self.flags = 4

        if self.italicAngle != 0:
            self.flags = self.flags | 64
        if usWeightClass >= 600:
            self.flags = self.flags | 262144
        if isFixedPitch:
            self.flags = self.flags | 1

        #################/
        # hhea - Horizontal header table
        #################/
        self.seek_table("hhea")
        self.skip(32)
        metricDataFormat = self.read_ushort()
        if metricDataFormat != 0:
            die("Unknown horizontal metric data format ".metricDataFormat)
        numberOfHMetrics = self.read_ushort()
        if numberOfHMetrics == 0:
            die("Number of horizontal metrics is 0")

        #################/
        # maxp - Maximum profile table
        #################/
        self.seek_table("maxp")
        self.skip(4)
        numGlyphs = self.read_ushort()

        #################/
        # cmap - Character to glyph index mapping table
        #################/
        cmap_offset = self.seek_table("cmap")
        self.skip(2)
        cmapTableCount = self.read_ushort()
        unicode_cmap_offset = 0
        unicode_cmap_offset12 = 0

        for i in range(cmapTableCount):
            platformID = self.read_ushort()
            encodingID = self.read_ushort()
            offset = self.read_ulong()
            save_pos = self._pos
            if platformID == 3 and encodingID == 10:  # Microsoft, UCS-4
                format = self.get_ushort(cmap_offset + offset)
                if format == 12:
                    if not unicode_cmap_offset12:
                        unicode_cmap_offset12 = cmap_offset + offset
                    break
            if (
                platformID == 3 and encodingID == 1
            ) or platformID == 0:  # Microsoft, Unicode
                format = self.get_ushort(cmap_offset + offset)
                if format == 4:
                    if not unicode_cmap_offset:
                        unicode_cmap_offset = cmap_offset + offset
                    break

            self.seek(save_pos)

        if not unicode_cmap_offset and not unicode_cmap_offset12:
            die(
                "Font ("
                + self.filename
                + ") does not have cmap for Unicode (platform 3, encoding 1, format 4, or platform 3, encoding 10, format 12, or platform 0, any encoding, format 4)"
            )

        glyphToChar = {}
        charToGlyph = {}
        if unicode_cmap_offset12:
            self.getCMAP12(unicode_cmap_offset12, glyphToChar, charToGlyph)
        else:
            self.getCMAP4(unicode_cmap_offset, glyphToChar, charToGlyph)

        #################/
        # hmtx - Horizontal metrics table
        #################/
        self.getHMTX(numberOfHMetrics, numGlyphs, glyphToChar, scale)

    ############################################/
    ############################################/

    def makeSubset(self, file, subset):
        self.filename = file
        with open(file, "rb") as self.fh:
            self._pos = 0
            self.charWidths = []
            self.glyphPos = {}
            self.charToGlyph = {}
            self.tables = {}
            self.otables = {}
            self.ascent = 0
            self.descent = 0
            self.skip(4)
            self.maxUni = 0
            self.readTableDirectory()

            #################/
            # head - Font header table
            #################/
            self.seek_table("head")
            self.skip(50)
            indexToLocFormat = self.read_ushort()
            glyphDataFormat = self.read_ushort()

            #################/
            # hhea - Horizontal header table
            #################/
            self.seek_table("hhea")
            self.skip(32)
            metricDataFormat = self.read_ushort()
            orignHmetrics = numberOfHMetrics = self.read_ushort()

            #################/
            # maxp - Maximum profile table
            #################/
            self.seek_table("maxp")
            self.skip(4)
            numGlyphs = self.read_ushort()

            #################/
            # cmap - Character to glyph index mapping table
            #################/
            cmap_offset = self.seek_table("cmap")
            self.skip(2)
            cmapTableCount = self.read_ushort()
            unicode_cmap_offset = 0
            unicode_cmap_offset12 = 0
            for i in range(cmapTableCount):
                platformID = self.read_ushort()
                encodingID = self.read_ushort()
                offset = self.read_ulong()
                save_pos = self._pos
                if platformID == 3 and encodingID == 10:  # Microsoft, UCS-4
                    format = self.get_ushort(cmap_offset + offset)
                    if format == 12:
                        if not unicode_cmap_offset12:
                            unicode_cmap_offset12 = cmap_offset + offset
                        break
                if (
                    platformID == 3 and encodingID == 1
                ) or platformID == 0:  # Microsoft, Unicode
                    format = self.get_ushort(cmap_offset + offset)
                    if format == 4:
                        unicode_cmap_offset = cmap_offset + offset
                        break

                self.seek(save_pos)

            if not unicode_cmap_offset and not unicode_cmap_offset12:
                die(
                    "Font ("
                    + self.filename
                    + ") does not have cmap for Unicode (platform 3, encoding 1, format 4, or platform 3, encoding 10, format 12, or platform 0, any encoding, format 4)"
                )

            glyphToChar = {}
            charToGlyph = {}
            if unicode_cmap_offset12:
                self.getCMAP12(unicode_cmap_offset12, glyphToChar, charToGlyph)
            else:
                self.getCMAP4(unicode_cmap_offset, glyphToChar, charToGlyph)

            self.charToGlyph = charToGlyph

            #################/
            # hmtx - Horizontal metrics table
            #################/
            scale = 1  # not used
            self.getHMTX(numberOfHMetrics, numGlyphs, glyphToChar, scale)

            #################/
            # loca - Index to location
            #################/
            self.getLOCA(indexToLocFormat, numGlyphs)

            subsetglyphs = [(0, 0)]  # special "sorted dict"!
            subsetCharToGlyph = {}
            for code in subset:
                if code in self.charToGlyph:
                    if (self.charToGlyph[code], code) not in subsetglyphs:
                        subsetglyphs.append(
                            (self.charToGlyph[code], code)
                        )  # Old Glyph ID => Unicode
                    subsetCharToGlyph[code] = self.charToGlyph[
                        code
                    ]  # Unicode to old GlyphID
                self.maxUni = max(self.maxUni, code)
            (start, dummy) = self.get_table_pos("glyf")

            subsetglyphs.sort()
            glyphSet = {}
            n = 0
            fsLastCharIndex = 0  # maximum Unicode index (character code) in this font, according to the cmap subtable for platform ID 3 and platform- specific encoding ID 0 or 1.
            for originalGlyphIdx, uni in subsetglyphs:
                fsLastCharIndex = max(fsLastCharIndex, uni)
                glyphSet[originalGlyphIdx] = n  # old glyphID to new glyphID
                n += 1

            codeToGlyph = {}
            for uni, originalGlyphIdx in sorted(subsetCharToGlyph.items()):
                codeToGlyph[uni] = glyphSet[originalGlyphIdx]

            self.codeToGlyph = codeToGlyph

            for originalGlyphIdx, uni in subsetglyphs:
                nonlocals = {
                    "start": start,
                    "glyphSet": glyphSet,
                    "subsetglyphs": subsetglyphs,
                }
                self.getGlyphs(originalGlyphIdx, nonlocals)

            numGlyphs = numberOfHMetrics = len(subsetglyphs)

            # tables copied from the original
            tags = ["name"]
            for tag in tags:
                self.add(tag, self.get_table(tag))
            tags = ["cvt ", "fpgm", "prep", "gasp"]
            for tag in tags:
                if tag in self.tables:
                    self.add(tag, self.get_table(tag))

            # post - PostScript
            opost = self.get_table("post")
            post = (
                b("\x00\x03\x00\x00")
                + substr(opost, 4, 12)
                + b("\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
            )
            self.add("post", post)

            # Sort CID2GID map into segments of contiguous codes
            if 0 in codeToGlyph:
                del codeToGlyph[0]
            # unset(codeToGlyph[65535])
            rangeid = 0
            range_ = {}
            prevcid = -2
            prevglidx = -1
            # for each character
            for cid, glidx in sorted(codeToGlyph.items()):
                if cid == (prevcid + 1) and glidx == (prevglidx + 1):
                    range_[rangeid].append(glidx)
                else:
                    # new range
                    rangeid = cid
                    range_[rangeid] = []
                    range_[rangeid].append(glidx)
                prevcid = cid
                prevglidx = glidx

            # cmap - Character to glyph mapping - Format 4 (MS / )
            segCount = len(range_) + 1  # + 1 Last segment has missing character 0xFFFF
            searchRange = 1
            entrySelector = 0
            while searchRange * 2 <= segCount:
                searchRange = searchRange * 2
                entrySelector = entrySelector + 1

            searchRange = searchRange * 2
            rangeShift = segCount * 2 - searchRange
            length = 16 + (8 * segCount) + (numGlyphs + 1)
            cmap = [
                0,
                1,  # Index : version, number of encoding subtables
                3,
                1,  # Encoding Subtable : platform (MS=3), encoding (Unicode)
                0,
                12,  # Encoding Subtable : offset (hi,lo)
                4,
                length,
                0,  # Format 4 Mapping subtable: format, length, language
                segCount * 2,
                searchRange,
                entrySelector,
                rangeShift,
            ]

            range_ = sorted(range_.items())

            # endCode(s)
            for start, subrange in range_:
                endCode = start + (len(subrange) - 1)
                cmap.append(endCode)  # endCode(s)

            cmap.append(0xFFFF)  # endCode of last Segment
            cmap.append(0)  # reservedPad

            # startCode(s)
            for start, subrange in range_:
                cmap.append(start)  # startCode(s)

            cmap.append(0xFFFF)  # startCode of last Segment
            # idDelta(s)
            for start, subrange in range_:
                idDelta = -(start - subrange[0])
                n += count(subrange)
                cmap.append(idDelta)  # idDelta(s)

            cmap.append(1)  # idDelta of last Segment
            # idRangeOffset(s)
            for subrange in range_:
                cmap.append(
                    0
                )  # idRangeOffset[segCount]      Offset in bytes to glyph indexArray, or 0

            cmap.append(0)  # idRangeOffset of last Segment
            for subrange, glidx in range_:
                cmap.extend(glidx)

            cmap.append(0)  # Mapping for last character
            cmapstr = b("")
            for cm in cmap:
                if cm >= 0:
                    cmapstr += pack(">H", cm)
                else:
                    try:
                        cmapstr += pack(">h", cm)
                    except:
                        warnings.warn("cmap value too big/small: %s" % cm)
                        cmapstr += pack(">H", -cm)
            self.add("cmap", cmapstr)

            # glyf - Glyph data
            (glyfOffset, glyfLength) = self.get_table_pos("glyf")
            if glyfLength < self.maxStrLenRead:
                glyphData = self.get_table("glyf")

            offsets = []
            glyf = b("")
            pos = 0

            hmtxstr = b("")
            xMinT = 0
            yMinT = 0
            xMaxT = 0
            yMaxT = 0
            advanceWidthMax = 0
            minLeftSideBearing = 0
            minRightSideBearing = 0
            xMaxExtent = 0
            maxPoints = 0  # points in non-compound glyph
            maxContours = 0  # contours in non-compound glyph
            maxComponentPoints = 0  # points in compound glyph
            maxComponentContours = 0  # contours in compound glyph
            maxComponentElements = 0  # number of glyphs referenced at top level
            maxComponentDepth = (
                0  # levels of recursion, set to 0 if font has only simple glyphs
            )
            self.glyphdata = {}

            for originalGlyphIdx, uni in subsetglyphs:
                # hmtx - Horizontal Metrics
                hm = self.getHMetric(orignHmetrics, originalGlyphIdx)
                hmtxstr += hm

                offsets.append(pos)
                try:
                    glyphPos = self.glyphPos[originalGlyphIdx]
                    glyphLen = self.glyphPos[originalGlyphIdx + 1] - glyphPos
                except IndexError:
                    warnings.warn("missing glyph %s" % (originalGlyphIdx))
                    glyphLen = 0

                if glyfLength < self.maxStrLenRead:
                    data = substr(glyphData, glyphPos, glyphLen)
                else:
                    if glyphLen > 0:
                        data = self.get_chunk(glyfOffset + glyphPos, glyphLen)
                    else:
                        data = b("")

                if glyphLen > 0:
                    up = unpack(">H", substr(data, 0, 2))[0]
                if glyphLen > 2 and (
                    up & (1 << 15)
                ):  # If number of contours <= -1 i.e. composite glyph
                    pos_in_glyph = 10
                    flags = GF_MORE
                    nComponentElements = 0
                    while flags & GF_MORE:
                        nComponentElements += (
                            1  # number of glyphs referenced at top level
                        )
                        up = unpack(">H", substr(data, pos_in_glyph, 2))
                        flags = up[0]
                        up = unpack(">H", substr(data, pos_in_glyph + 2, 2))
                        glyphIdx = up[0]
                        self.glyphdata.setdefault(originalGlyphIdx, {}).setdefault(
                            "compGlyphs", []
                        ).append(glyphIdx)
                        try:
                            data = self._set_ushort(
                                data, pos_in_glyph + 2, glyphSet[glyphIdx]
                            )
                        except KeyError:
                            data = 0
                            warnings.warn("missing glyph data %s" % glyphIdx)
                        pos_in_glyph += 4
                        if flags & GF_WORDS:
                            pos_in_glyph += 4
                        else:
                            pos_in_glyph += 2
                        if flags & GF_SCALE:
                            pos_in_glyph += 2
                        elif flags & GF_XYSCALE:
                            pos_in_glyph += 4
                        elif flags & GF_TWOBYTWO:
                            pos_in_glyph += 8

                    maxComponentElements = max(maxComponentElements, nComponentElements)

                glyf += data
                pos += glyphLen
                if pos % 4 != 0:
                    padding = 4 - (pos % 4)
                    glyf += str_repeat(b("\0"), padding)
                    pos += padding

            offsets.append(pos)
            self.add("glyf", glyf)

            # hmtx - Horizontal Metrics
            self.add("hmtx", hmtxstr)

            # loca - Index to location
            locastr = b("")
            if ((pos + 1) >> 1) > 0xFFFF:
                indexToLocFormat = 1  # long format
                for offset in offsets:
                    locastr += pack(">L", offset)
            else:
                indexToLocFormat = 0  # short format
                for offset in offsets:
                    locastr += pack(">H", offset // 2)

            self.add("loca", locastr)

            # head - Font header
            head = self.get_table("head")
            head = self._set_ushort(head, 50, indexToLocFormat)
            self.add("head", head)

            # hhea - Horizontal Header
            hhea = self.get_table("hhea")
            hhea = self._set_ushort(hhea, 34, numberOfHMetrics)
            self.add("hhea", hhea)

            # maxp - Maximum Profile
            maxp = self.get_table("maxp")
            maxp = self._set_ushort(maxp, 4, numGlyphs)
            self.add("maxp", maxp)

            # OS/2 - OS/2
            os2 = self.get_table("OS/2")
            self.add("OS/2", os2)

        # Put the TTF file together
        stm = self.endTTFile("")
        return stm

    #########################################
    # Recursively get composite glyph data
    def getGlyphData(self, originalGlyphIdx, nonlocals):
        # &maxdepth, &depth, &points, &contours
        nonlocals["depth"] += 1
        nonlocals["maxdepth"] = max(nonlocals["maxdepth"], nonlocals["depth"])
        if len(self.glyphdata[originalGlyphIdx]["compGlyphs"]):
            for glyphIdx in self.glyphdata[originalGlyphIdx]["compGlyphs"]:
                self.getGlyphData(glyphIdx, nonlocals)

        elif (self.glyphdata[originalGlyphIdx]["nContours"] > 0) and nonlocals[
            "depth"
        ] > 0:  # simple
            contours += self.glyphdata[originalGlyphIdx]["nContours"]
            points += self.glyphdata[originalGlyphIdx]["nPoints"]

        nonlocals["depth"] -= 1

    #########################################
    # Recursively get composite glyphs
    def getGlyphs(self, originalGlyphIdx, nonlocals):
        # &start, &glyphSet, &subsetglyphs)

        try:
            glyphPos = self.glyphPos[originalGlyphIdx]
            glyphLen = self.glyphPos[originalGlyphIdx + 1] - glyphPos
        except IndexError:
            warnings.warn("missing glyph %s" % (originalGlyphIdx))
            return

        if not glyphLen:
            return

        self.seek(nonlocals["start"] + glyphPos)
        numberOfContours = self.read_short()
        if numberOfContours < 0:
            self.skip(8)
            flags = GF_MORE
            while flags & GF_MORE:
                flags = self.read_ushort()
                glyphIdx = self.read_ushort()
                if glyphIdx not in nonlocals["glyphSet"]:
                    nonlocals["glyphSet"][glyphIdx] = len(
                        nonlocals["subsetglyphs"]
                    )  # old glyphID to new glyphID
                    nonlocals["subsetglyphs"].append((glyphIdx, 1))

                savepos = self.fh.tell()
                self.getGlyphs(glyphIdx, nonlocals)
                self.seek(savepos)
                if flags & GF_WORDS:
                    self.skip(4)
                else:
                    self.skip(2)
                if flags & GF_SCALE:
                    self.skip(2)
                elif flags & GF_XYSCALE:
                    self.skip(4)
                elif flags & GF_TWOBYTWO:
                    self.skip(8)

    #########################################

    def getHMTX(self, numberOfHMetrics, numGlyphs, glyphToChar, scale):
        start = self.seek_table("hmtx")
        aw = 0
        self.charWidths = []

        def resize_cw(size, default):
            size = (((size + 1) // 1024) + 1) * 1024
            delta = size - len(self.charWidths)
            if delta > 0:
                self.charWidths += [default] * delta

        nCharWidths = 0
        if (numberOfHMetrics * 4) < self.maxStrLenRead:
            data = self.get_chunk(start, (numberOfHMetrics * 4))
            arr = unpack(">%dH" % (len(data) // 2), data)
        else:
            self.seek(start)
        for glyph in range(numberOfHMetrics):
            if (numberOfHMetrics * 4) < self.maxStrLenRead:
                aw = arr[(glyph * 2)]  # PHP starts arrays from index 0!? +1
            else:
                aw = self.read_ushort()
                lsb = self.read_ushort()

            if glyph in glyphToChar or glyph == 0:
                if aw >= (1 << 15):
                    aw = 0  # 1.03 Some (arabic) fonts have -ve values for width
                    # although should be unsigned value - comes out as e.g. 65108 (intended -50)
                if glyph == 0:
                    self.defaultWidth = scale * aw
                    continue

                for char in glyphToChar[glyph]:
                    if char != 0 and char != 65535:
                        w = int(
                            round(scale * aw + 0.001)
                        )  # ROUND_HALF_UP in PY3K (like php)
                        if w == 0:
                            w = 65535
                        if char < 196608:
                            if char >= len(self.charWidths):
                                resize_cw(char, self.defaultWidth)
                            self.charWidths[char] = w
                            nCharWidths += 1

        data = self.get_chunk((start + numberOfHMetrics * 4), (numGlyphs * 2))
        arr = unpack(">%dH" % (len(data) // 2), data)
        diff = numGlyphs - numberOfHMetrics
        for pos in range(diff):
            glyph = pos + numberOfHMetrics
            if glyph in glyphToChar:
                for char in glyphToChar[glyph]:
                    if char != 0 and char != 65535:
                        w = int(
                            round(scale * aw + 0.001)
                        )  # ROUND_HALF_UP in PY3K (like php)
                        if w == 0:
                            w = 65535
                        if char < 196608:
                            if char >= len(self.charWidths):
                                resize_cw(char, self.defaultWidth)
                            self.charWidths[char] = w
                            nCharWidths += 1

        # NB 65535 is a set width of 0
        # First bytes define number of chars in font
        self.charWidths[0] = nCharWidths

    def getHMetric(self, numberOfHMetrics, gid):
        start = self.seek_table("hmtx")
        if gid < numberOfHMetrics:
            self.seek(start + (gid * 4))
            hm = self.fh.read(4)
        else:
            self.seek(start + ((numberOfHMetrics - 1) * 4))
            hm = self.fh.read(2)
            self.seek(start + (numberOfHMetrics * 2) + (gid * 2))
            hm += self.fh.read(2)
        return hm

    def getLOCA(self, indexToLocFormat, numGlyphs):
        start = self.seek_table("loca")
        self.glyphPos = []
        if indexToLocFormat == 0:
            data = self.get_chunk(start, (numGlyphs * 2) + 2)
            arr = unpack(">%dH" % (len(data) // 2), data)
            for n in range(numGlyphs):
                self.glyphPos.append((arr[n] * 2))  # n+1 !?
        elif indexToLocFormat == 1:
            data = self.get_chunk(start, (numGlyphs * 4) + 4)
            arr = unpack(">%dL" % (len(data) // 4), data)
            for n in range(numGlyphs):
                self.glyphPos.append((arr[n]))  # n+1 !?
        else:
            die("Unknown location table format " + indexToLocFormat)

    # CMAP Format 4
    def getCMAP4(self, unicode_cmap_offset, glyphToChar, charToGlyph):
        self.maxUniChar = 0
        self.seek(unicode_cmap_offset + 2)
        length = self.read_ushort()
        limit = unicode_cmap_offset + length
        self.skip(2)

        segCount = self.read_ushort() // 2
        self.skip(6)
        endCount = []
        for i in range(segCount):
            endCount.append(self.read_ushort())
        self.skip(2)
        startCount = []
        for i in range(segCount):
            startCount.append(self.read_ushort())
        idDelta = []
        for i in range(segCount):
            idDelta.append(self.read_short())  # ???? was unsigned short
        idRangeOffset_start = self._pos
        idRangeOffset = []
        for i in range(segCount):
            idRangeOffset.append(self.read_ushort())

        for n in range(segCount):
            endpoint = endCount[n] + 1
            for unichar in range(startCount[n], endpoint, 1):
                if idRangeOffset[n] == 0:
                    glyph = (unichar + idDelta[n]) & 0xFFFF
                else:
                    offset = (unichar - startCount[n]) * 2 + idRangeOffset[n]
                    offset = idRangeOffset_start + 2 * n + offset
                    if offset >= limit:
                        glyph = 0
                    else:
                        glyph = self.get_ushort(offset)
                        if glyph != 0:
                            glyph = (glyph + idDelta[n]) & 0xFFFF

                charToGlyph[unichar] = glyph
                if unichar < 196608:
                    self.maxUniChar = max(unichar, self.maxUniChar)
                glyphToChar.setdefault(glyph, []).append(unichar)

    # CMAP Format 12
    def getCMAP12(self, unicode_cmap_offset, glyphToChar, charToGlyph):
        self.maxUniChar = 0
        # table (skip format version, should be 12)
        self.seek(unicode_cmap_offset + 2)
        # reserved
        self.skip(2)
        # table length
        length = self.read_ulong()
        # language (should be 0)
        self.skip(4)
        # groups count
        grpCount = self.read_ulong()

        if 2 + 2 + 4 + 4 + 4 + grpCount * 3 * 4 > length:
            die("TTF format 12 cmap table too small")
        for n in range(grpCount):
            startCharCode = self.read_ulong()
            endCharCode = self.read_ulong()
            glyph = self.read_ulong()
            for unichar in range(startCharCode, endCharCode + 1):
                charToGlyph[unichar] = glyph
                if unichar < 196608:
                    self.maxUniChar = max(unichar, self.maxUniChar)
                glyphToChar.setdefault(glyph, []).append(unichar)
                glyph += 1

    # Put the TTF file together
    def endTTFile(self, stm):
        stm = b("")
        numTables = count(self.otables)
        searchRange = 1
        entrySelector = 0
        while searchRange * 2 <= numTables:
            searchRange = searchRange * 2
            entrySelector = entrySelector + 1

        searchRange = searchRange * 16
        rangeShift = numTables * 16 - searchRange

        # Header
        if _TTF_MAC_HEADER:
            stm += pack(
                ">LHHHH", 0x74727565, numTables, searchRange, entrySelector, rangeShift
            )  # Mac
        else:
            stm += pack(
                ">LHHHH", 0x00010000, numTables, searchRange, entrySelector, rangeShift
            )  # Windows

        # Table directory
        tables = self.otables

        offset = 12 + numTables * 16
        sorted_tables = sorted(tables.items())
        for tag, data in sorted_tables:
            if tag == "head":
                head_start = offset
            stm += tag.encode("latin1")
            checksum = calcChecksum(data)
            stm += pack(">HH", checksum[0], checksum[1])
            stm += pack(">LL", offset, strlen(data))
            paddedLength = (strlen(data) + 3) & ~3
            offset = offset + paddedLength

        # Table data
        for tag, data in sorted_tables:
            data += b("\0\0\0")
            stm += substr(data, 0, (strlen(data) & ~3))

        checksum = calcChecksum(stm)
        checksum = sub32((0xB1B0, 0xAFBA), checksum)
        chk = pack(">HH", checksum[0], checksum[1])
        stm = self.splice(stm, (head_start + 8), chk)
        return stm
